@@ -6,7 +6,7 @@ import django
 from django.conf import settings
 from django.test.utils import get_runner
 
-os.environ['DJANGO_SETTINGS_MODULE'] = 'fallball.settings'
+os.environ['DJANGO_SETTINGS_MODULE'] = 'fallballapp.tests.settings'
 django.setup()
 
 TestRunner = get_runner(settings)
