@@ -1,12 +1,13 @@
 import pkg_resources
 
-from django.contrib.auth.models import User
+from django.contrib.auth import get_user_model
 from django.conf import settings
 from django.shortcuts import get_object_or_404
 from rest_framework import status
 from rest_framework.authentication import TokenAuthentication, BasicAuthentication
-from rest_framework.decorators import detail_route, list_route
+from rest_framework.decorators import detail_route
 from rest_framework.permissions import IsAuthenticated
+from rest_framework.renderers import JSONRenderer
 from rest_framework.response import Response
 from rest_framework.status import HTTP_404_NOT_FOUND, HTTP_400_BAD_REQUEST
 from rest_framework.views import APIView
@@ -21,19 +22,19 @@ except ImportError:
     from urllib import urlencode
 
 from fallballapp.middleware import logger
-
-from fallballapp.models import Application, Client, ClientUser, Reseller
-from fallballapp.serializers import (ApplicationSerializer, ClientSerializer,
-                                     ClientUserSerializer, ResellerSerializer,
+from fallballapp.models import Application, Client, ClientUser, Reseller, UNLIMITED
+from fallballapp.renderers import PlainTextRenderer
+from fallballapp.serializers import (ApplicationSerializer, ApplicationPutSerializer,
+                                     ClientSerializer, ClientUserSerializer, ResellerSerializer,
                                      ResellerNameSerializer, UserAuthorizationSerializer)
 from fallballapp.utils import (get_app_username, get_object_or_403, get_jwt_token,
                                is_superuser, is_application, get_user_context, free_space)
 
 
 class Description(APIView):
-    def get(self, request, format=None):
+    def get(self, *args, **kwargs):
         env = pkg_resources.Environment()
-        res = env._distmap.get('fallball-connector', [None])[0]
+        res = env._distmap.get('fallball', [None])[0]
         version = res.version if res else ''
         return Response({'description': 'Fallball - File sharing, that everyone learns',
                          'version': version})
@@ -42,9 +43,13 @@ class Description(APIView):
 class ApplicationViewSet(ModelViewSet):
     authentication_classes = (TokenAuthentication,)
     permission_classes = (IsAuthenticated,)
-    serializer_class = ApplicationSerializer
     queryset = Application.objects.all()
 
+    def get_serializer_class(self):
+        if self.request.method == 'PUT':
+            return ApplicationPutSerializer
+        return ApplicationSerializer
+
     @is_superuser
     def create(self, request, *args, **kwargs):
         return ModelViewSet.create(self, request, *args, **kwargs)
@@ -61,6 +66,10 @@ class ApplicationViewSet(ModelViewSet):
     def destroy(self, request, *args, **kwargs):
         return ModelViewSet.destroy(self, request, *args, **kwargs)
 
+    @is_superuser
+    def update(self, request, *args, **kwargs):
+        return ModelViewSet.update(self, request, *args, partial=True, **kwargs)
+
 
 class ResellerViewSet(ModelViewSet):
     authentication_classes = (TokenAuthentication, JSONWebTokenAuthentication)
@@ -78,8 +87,6 @@ class ResellerViewSet(ModelViewSet):
     def destroy(self, request, *args, **kwargs):
         Reseller.objects.filter(name=kwargs['name'],
                                 application=kwargs['application']).delete()
-        username = get_app_username(kwargs['name'], kwargs['application'])
-        User.objects.filter(username=username).delete()
         return Response('Reseller has been deleted', status=status.HTTP_204_NO_CONTENT)
 
     def list(self, request, *args, **kwargs):
@@ -89,12 +96,12 @@ class ResellerViewSet(ModelViewSet):
         else:
             resellers = Reseller.objects.filter(owner=request.user)
             if not resellers:
-                admin = ClientUser.objects.filter(owner=request.user, admin=True).first()
-                if not admin:
+                client_user = ClientUser.objects.filter(owner=request.user).first()
+                if not client_user:
                     return Response("Resellers do not exist for such account",
                                     status=HTTP_404_NOT_FOUND)
 
-                queryset = [admin.client.reseller, ]
+                queryset = [client_user.client.reseller, ]
                 serializer = ResellerNameSerializer(queryset, many=True)
                 return Response(serializer.data)
 
@@ -114,20 +121,6 @@ class ResellerViewSet(ModelViewSet):
         serializer = ResellerSerializer(queryset, many=True)
         return Response(serializer.data[0])
 
-    @detail_route(methods=['get'])
-    def reset(self, request, *args, **kwargs):
-        """
-        Repair particular reseller
-        """
-        return Response("Method is not implemented yet", status=status.HTTP_405_METHOD_NOT_ALLOWED)
-
-    @list_route(methods=['get'])
-    def reset_all(self, request, *args, **kwargs):
-        """
-        Repair all resellers
-        """
-        return Response("Method is not implemented yet", status=status.HTTP_405_METHOD_NOT_ALLOWED)
-
 
 class ClientViewSet(ModelViewSet):
     """
@@ -152,9 +145,10 @@ class ClientViewSet(ModelViewSet):
                                          owner=request.user)
         if not Client.objects.filter(reseller=reseller,
                                      name=request.data['name']):
-            # Check if there is a free space for new client
-            free_space = reseller.limit - reseller.get_usage()
-            if free_space >= request.data['storage']['limit']:
+
+            # Check if there is a free space for new client or storage is unlimited
+            space = free_space(reseller)
+            if space >= request.data['storage']['limit'] or reseller.limit is UNLIMITED:
                 # Every client should belong to particular reseller
                 request.data['reseller'] = reseller
                 return ModelViewSet.create(self, request, *args, **kwargs)
@@ -219,22 +213,9 @@ class ClientViewSet(ModelViewSet):
             return Response('Client has been deleted', status=status.HTTP_204_NO_CONTENT)
         return Response('Such client does not exist', status=status.HTTP_400_BAD_REQUEST)
 
-    @detail_route(methods=['get'])
-    def reset(self, request, *args, **kwargs):
-        """
-        Recreate client to initial state
-        """
-        return Response("Method is not implemented yet", status=status.HTTP_405_METHOD_NOT_ALLOWED)
-
-    @list_route(methods=['get'])
-    def reset_all(self, request, *args, **kwargs):
-        """
-        Recreate all reseller clients to initial state
-        """
-        return Response("Method is not implemented yet", status=status.HTTP_405_METHOD_NOT_ALLOWED)
-
 
 class ClientUserViewSet(ModelViewSet):
+    renderer_classes = (JSONRenderer, PlainTextRenderer,)
     queryset = ClientUser.objects.all().order_by('-id')
     serializer_class = ClientUserSerializer
     authentication_classes = (TokenAuthentication, JSONWebTokenAuthentication)
@@ -255,13 +236,18 @@ class ClientUserViewSet(ModelViewSet):
         if 'usage' in request.data:
             return Response("Usage should not be specified", status=status.HTTP_400_BAD_REQUEST)
 
-        # Check if client has free space for new user
+        # Check if client has free space for new user or storage is unlimited
         space = free_space(client)
-        if space >= request.data['storage']['limit']:
+        if space >= request.data['storage']['limit'] or space is UNLIMITED:
             request.data['client'] = client
             request.data['application_id'] = reseller.application.id
             if 'admin' not in request.data:
                 request.data['admin'] = False
+            if 'password' in request.data:
+                request.data.pop('password')
+            if 'profile_type' in request.data:
+                if request.data['profile_type'] not in dict(ClientUser.USER_PROFILE_TYPES):
+                    request.data['profile_type'] = ClientUser.DEFAULT_PROFILE
 
             return ModelViewSet.create(self, *args, **kwargs)
 
@@ -271,8 +257,6 @@ class ClientUserViewSet(ModelViewSet):
     def destroy(self, *args, **kwargs):
         client_user = get_object_or_404(ClientUser, email=kwargs['email'], client=kwargs['client'])
         client_user.delete()
-        User.objects.filter(username='{}.{}'.format(kwargs['reseller'].application.id,
-                                                    kwargs['email'])).delete()
         return Response("User has been deleted", status=status.HTTP_204_NO_CONTENT)
 
     @get_user_context
@@ -299,37 +283,41 @@ class ClientUserViewSet(ModelViewSet):
             return self.create(*args, **kwargs)
 
         request = args[0]
-        if (all(field in request.data for field in ('password', 'storage')) and
-                all(field in request.data['storage'] for field in ('usage', 'limit'))):
 
-            limit = request.data['storage']['limit']
-            usage = request.data['storage']['usage']
+        if 'password' in request.data:
+            user = get_object_or_404(get_user_model(),
+                                     username=get_app_username(kwargs['application'],
+                                                               kwargs['email']))
+            user.set_password(request.data['password'])
+            user.save()
+            client_user.password = request.data['password']
+
+        if 'storage' in request.data:
+            limit = request.data['storage'].get('limit')
+            usage = request.data['storage'].get('usage')
+
+            if limit is not None:
+                client_user.limit = limit
+            if usage is not None:
+                client_user.usage = usage
 
             space = free_space(kwargs['client'])
-            if space < limit or usage > limit:
+            if space < client_user.limit or client_user.usage > client_user.limit:
                 return Response("Storage limit is reached", HTTP_400_BAD_REQUEST)
 
-            user = get_object_or_404(User, username=get_app_username(kwargs['application'],
-                                                                     kwargs['email']))
-            user.set_password(request.data['password'])
+        if 'admin' in request.data:
+            client_user.admin = request.data['admin']
 
-            client_user = get_object_or_404(ClientUser, client=kwargs['client'],
-                                            email=kwargs['email'])
-            client_user.password = request.data['password']
-            client_user.usage = request.data['storage']['usage']
-            if 'admin' in request.data:
-                client_user.admin = True
-            client_user.limit = request.data['storage']['limit']
+        if 'profile_type' in request.data:
+            if request.data['profile_type'] in dict(ClientUser.USER_PROFILE_TYPES):
+                client_user.profile_type = request.data['profile_type']
 
-            user.save()
-            client_user.save()
+        client_user.save()
 
-            queryset = [client_user, ]
-            serializer = ClientUserSerializer(queryset, many=True)
+        queryset = [client_user, ]
+        serializer = ClientUserSerializer(queryset, many=True)
 
-            return Response(serializer.data[0], status=status.HTTP_200_OK)
-
-        return Response("Wrong parameters", HTTP_400_BAD_REQUEST)
+        return Response(serializer.data[0], status=status.HTTP_200_OK)
 
     @detail_route(methods=['get'])
     def token(self, request, **kwargs):
@@ -365,7 +353,10 @@ class ClientUserViewSet(ModelViewSet):
             reseller = Reseller.objects.filter(name=kwargs['reseller_name'],
                                                owner=request.user).first()
             if not reseller:
-                return Response("Could not determine application", status=HTTP_404_NOT_FOUND)
+                admin = ClientUser.objects.filter(owner=request.user, admin=True).first()
+                if not admin:
+                    return Response("Could not determine application", status=HTTP_404_NOT_FOUND)
+                reseller = admin.client.reseller
 
             application_id = reseller.application.id
 
